@@ -18,6 +18,7 @@ import Data.Map.Strict qualified as M
 import Data.Text qualified as T
 import Data.Text.Encoding qualified as T
 import Data.Text.IO qualified as T
+import Data.Time.Clock.System (SystemTime (..), getSystemTime)
 import Futhark.Analysis.Metrics.Type
 import Futhark.Server
 import Futhark.Test
@@ -36,6 +37,10 @@ import Text.Regex.TDFA
 
 --- Test execution
 
+-- The use of [T.Text] here is somewhat kludgy. We use it to track how
+-- many errors have occurred during testing of a single program (which
+-- may have multiple entry points). This should really not be done at
+-- the monadic level - a test failing should be handled explicitly.
 type TestM = ExceptT [T.Text] IO
 
 -- Taken from transformers-0.5.5.0.
@@ -113,6 +118,8 @@ data TestMode
     Compiled
   | -- | Test interpreted code.
     Interpreted
+  | -- | Perform structure tests.
+    Structure
   deriving (Eq, Show)
 
 data TestCase = TestCase
@@ -249,7 +256,7 @@ runTestCase (TestCase mode program testcase progs) = do
           ]
   case testAction testcase of
     CompileTimeFailure expected_error ->
-      unless (mode == Internalise) . context checkctx $ do
+      unless (mode `elem` [Structure, Internalise]) . context checkctx $ do
         (code, _, err) <-
           liftIO $ readProcessWithExitCode futhark ["check", program] ""
         case code of
@@ -283,12 +290,15 @@ runTestCase (TestCase mode program testcase progs) = do
         context "Generating reference outputs" $
           -- We probably get the concurrency at the test program level,
           -- so force just one data set at a time here.
-          ensureReferenceOutput (Just 1) (FutharkExe futhark) "c" program ios
+          withExceptT pure $
+            ensureReferenceOutput (Just 1) (FutharkExe futhark) "c" program ios
+
+      when (mode == Structure) $
+        mapM_ (testMetrics progs program) structures
 
       when (mode `elem` [Compile, Compiled]) $
         context ("Compiling with --backend=" <> T.pack backend) $ do
           compileTestProgram extra_compiler_options (FutharkExe futhark) backend program warnings
-          mapM_ (testMetrics progs program) structures
           unless (mode == Compile) $ do
             (tuning_opts, _) <-
               liftIO $ determineTuning (configTuning progs) program
@@ -388,7 +398,9 @@ runResult _ (ExitFailure _) _ stderr_s =
 
 compileTestProgram :: [String] -> FutharkExe -> String -> FilePath -> [WarningTest] -> TestM ()
 compileTestProgram extra_options futhark backend program warnings = do
-  (_, futerr) <- compileProgram ("--server" : extra_options) futhark backend program
+  (_, futerr) <-
+    withExceptT pure $
+      compileProgram ("--server" : extra_options) futhark backend program
   testWarnings warnings futerr
 
 compareResult ::
@@ -509,6 +521,26 @@ reportTable ts = do
     running = labelstr <> (T.unwords . reverse . map (T.pack . testCaseProgram) . testStatusRun) ts
     labelstr = "Now testing: "
 
+reportLine :: MVar SystemTime -> TestStatus -> IO ()
+reportLine time_mvar ts =
+  modifyMVar_ time_mvar $ \time -> do
+    time_now <- getSystemTime
+    if systemSeconds time_now - systemSeconds time >= period
+      then do
+        T.putStrLn $
+          "("
+            <> showText (testStatusFail ts)
+            <> " failed, "
+            <> showText (testStatusPass ts)
+            <> " passed, "
+            <> showText num_remain
+            <> " to go)."
+        pure time_now
+      else pure time
+  where
+    num_remain = length $ testStatusRemain ts
+    period = 60
+
 moveCursorToTableTop :: IO ()
 moveCursorToTableTop = cursorUpLine tableLines
 
@@ -531,11 +563,13 @@ runTests config paths = do
   let (excluded, included) = partition (excludedTest config) all_tests
   _ <- forkIO $ mapM_ (putMVar testmvar . excludeCases config) included
 
+  time_mvar <- newMVar $ MkSystemTime 0 0
+
   let fancy = not (configLineOutput config) && fancyTerminal
 
       report
         | fancy = reportTable
-        | otherwise = const (pure ())
+        | otherwise = reportLine time_mvar
       clear
         | fancy = clearFromCursorToScreenEnd
         | otherwise = pure ()
@@ -562,8 +596,7 @@ runTests config paths = do
                         { testStatusRemain = test `delete` testStatusRemain ts,
                           testStatusRun = test `delete` testStatusRun ts,
                           testStatusRunsRemain =
-                            testStatusRunsRemain ts
-                              - numTestCases test
+                            testStatusRunsRemain ts - numTestCases test
                         }
                 case res of
                   Success -> do
@@ -709,6 +742,11 @@ commandLineOptions =
       ["compile"]
       (NoArg $ Right $ \config -> config {configTestMode = Compile})
       "Only compile, do not run.",
+    Option
+      "s"
+      ["structure"]
+      (NoArg $ Right $ \config -> config {configTestMode = Structure})
+      "Perform structure tests.",
     Option
       "I"
       ["internalise"]
