@@ -196,7 +196,7 @@ partitionSeqDims (SegSeqDims seq_is) space =
 compileFlatId :: SegSpace -> InKernelGen ()
 compileFlatId space = do
   ltid <- kernelLocalThreadId . kernelConstants <$> askEnv
-  dPrimV_ (segFlat space) ltid
+  dPrimV_ (segFlat space) $ sExt64 ltid
 
 -- Construct the necessary lock arrays for an intra-block histogram.
 prepareIntraBlockSegHist ::
@@ -281,9 +281,9 @@ compileBlockExp (Pat [pe]) (BasicOp (Opaque _ se)) =
 compileBlockExp (Pat [dest]) (BasicOp (ArrayLit es _)) =
   forM_ (zip [0 ..] es) $ \(i, e) ->
     copyDWIMFix (patElemName dest) [fromIntegral (i :: Int64)] e []
-compileBlockExp _ (BasicOp (UpdateAcc acc is vs)) = do
+compileBlockExp _ (BasicOp (UpdateAcc safety acc is vs)) = do
   ltid <- kernelLocalThreadId . kernelConstants <$> askEnv
-  sWhen (ltid .==. 0) $ updateAcc acc is vs
+  sWhen (ltid .==. 0) $ updateAcc safety acc is vs
   sOp $ Imp.Barrier Imp.FenceLocal
 compileBlockExp (Pat [dest]) (BasicOp (Replicate ds se)) | ds /= mempty = do
   flat <- newVName "rep_flat"
@@ -602,7 +602,7 @@ sKernelBlock ::
   KernelAttrs ->
   InKernelGen () ->
   CallKernelGen ()
-sKernelBlock = sKernel blockOperations kernelBlockId
+sKernelBlock = sKernel blockOperations $ sExt64 . kernelBlockId
 
 compileBlockResult ::
   SegSpace ->
